/*
 *	Alpha-numeric-LCD-module driver
 *	for LCD controllers HD44780 and HD66712
 *	connected to PC compatible parallel port (see README)
 *
 *	Copyright 1998-2001 Nils Faerber <nils@kernelconcepts.de>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Changes:
 *
 *	1998		Initial release, Nils Faerber
 *
 *	1999		Fixes, updates for kernel 2.1/2.2, Nils Faerber
 *
 *	20000826	Added module parameters and defaults
 *
 *			Juergen Bauer
 *
 *	20001225	Updated driver to kernel 2.4, some cleanups/fixes
 *			Added proper printk KERN_* log-values
 *			Added new controller type (a_type==1)
 *
 *			Nils Faerber <nils@kernelconcepts.de>
 *
 *      2001
 *      2002
 *      20021225        Hacked the code to support setting the cursor key
 *                      by sending a newly invented ESC command to the
 *                      lcd device file (needed this hack to port CMMS
 *                      to this LCD driver):
 *                      ESC p-cc,l - where cc is column (two decimals) and
 *                      l is the line the cursor position will be set to
 *
 *                      Janosch Rolles <janosch@base2.de>
 *                      for the CMMS car mp3 player project from
 *                      http://cmms.sourceforge.net
 */

#include <linux/ioport.h>

#ifdef KERNEL20
#include <asm/segment.h>
#include <linux/sched.h>
#endif
#ifdef KERNEL21
#include <asm/uaccess.h>
#endif

#include <asm/io.h>
#include <linux/module.h>

#include <linux/fs.h>
#include <linux/delay.h>
#include <asm/delay.h>
/*#include <string.h>*/
#include "version.h"
#include "c_table.h"


/*
 * LCD pins and functions
 */
/* control bits */
#define LCD_RS		0x02
#define LCD_EN		0x01
/* functions (RS==1) */
#define LCD_CLEAR	0x01
#define LCD_HOME	0x02
#define LCD_MODESET	0x04
#define LCD_ONOFF	0x08
#define LCD_EXTFUNC	0x08	/* RE must be set to 1 using LCD_FUNCSET */
#define LCD_SHIFT	0x10
#define LCD_SCRL_EN	0x10	/* RE == 1 */
#define LCD_FUNCSET	0x20
#define LCD_SETCGRAM	0x40
#define LCD_SETSEGRAM	0x40	/* RE == 1 */
#define LCD_SETDDRAM	0x80
#define LCD_SETSCRL	0x80	/* RE == 1 */

/*
 * Some globals
 */
static int lcd_usage=0;
static char dline=0,dcol=0;

/*
 * Setting module defaults
 */
const int LCD_MAJOR=120;

static int LCD_ADDRESS=0x0378;
static int LCD_COLS=20;
static int LCD_LINES=4;
static int TIME_SHORT=40;
static int TIME_LONG=100;
  
static int io=0;
static int cols=0;
static int lines=0;
static int t_short=0;
static int t_long=0;
static int a_type=0;

#ifdef KERNEL21
MODULE_PARM(io, "i"); 
MODULE_PARM(cols, "i");
MODULE_PARM(lines, "i");
MODULE_PARM(t_short, "i");
MODULE_PARM(t_long, "i");
MODULE_PARM(a_type, "i");
#endif

/*
 * Some prototypes
 */
void pos_cursor(unsigned char, unsigned char);


/*
 * Sleep nicely for 'n' uS
 * Maybe this saves some cpu cycles and sysload
 */
static void sleep_delay(unsigned long n)
{
int d=n/(1000000/HZ);

	if (!d)
		udelay(n);
	else {
		/* Yield CPU time */
		unsigned long x=jiffies;
		while((jiffies-x) <= d)
			schedule();
	}
}


/*
 * LCD specific commands
 */

/*
 * Set to instruction mode and write a command byte
 */
void write_command(char command)
{
	outb(LCD_RS | LCD_EN, LCD_ADDRESS + 2);
	USLEEP(TIME_SHORT);
	outb(LCD_RS, LCD_ADDRESS + 2);
	USLEEP(TIME_SHORT);
	outb(command, LCD_ADDRESS);  
	USLEEP(TIME_SHORT);
	outb(LCD_RS | LCD_EN, LCD_ADDRESS + 2);
	USLEEP(TIME_SHORT);
	if (command == LCD_CLEAR) USLEEP(2000);	/* 1640 works on mine but others seem slower */
}


/*
 * Set data mode and write a data byte
 */
void write_data(char data)
{
	outb(0x00, LCD_ADDRESS + 2);	/* set EN to 1 and RS to 1 */
	USLEEP(TIME_SHORT);
	/* write the data translated by c_table.h */
	outb(c_table[(unsigned char)data], LCD_ADDRESS);
	USLEEP(TIME_LONG);
	outb(LCD_EN, LCD_ADDRESS + 2);	/* clear out EN */
	USLEEP(TIME_SHORT);
}


/*
 * Resets LCD panel and sets initial mode, i.e.
 * underline cursor, not blinking
 * one line mode
 * small font
 */
void init_lcd(void)
{
	USLEEP(15000);
	write_command(LCD_FUNCSET | 0x10);		/* Set 8-bit mode */
	USLEEP(4200);
	write_command(LCD_FUNCSET | 0x10);		/* Set 8-bit mode */
	USLEEP(200);
	write_command(LCD_FUNCSET | 0x10);		/* Set 8-bit mode */
	USLEEP(4200);
	write_command(LCD_FUNCSET | 0x10 | 0x08);	/* Set 8-bit mode, font, matrix */
	USLEEP(240);
	write_command(LCD_ONOFF);			/* Set display off */
	USLEEP(240);
	write_command(LCD_ONOFF | 0x04 | 0x02);	/* Set display on, cursor on */
	USLEEP(1740);
	write_command(LCD_MODESET | 0x02);		/* Set entry mode, cursor moves right, shift display */
	if ((a_type == 1) && (LCD_LINES > 2)) {
		write_command(LCD_FUNCSET | 0x10 | 0x08 | 0x04);	/* enter ext-func-set */
		USLEEP(240);
		write_command(LCD_EXTFUNC | 0x01);		/* Set 4-line mode */
		USLEEP(240);
		write_command(LCD_FUNCSET | 0x10 | 0x08);	/* leave ext-func-set */
		USLEEP(240);
	}
	USLEEP(240);
	write_command(LCD_CLEAR);
}


void init_cgram(void)
{
unsigned int a,b;

	write_command(LCD_SETCGRAM);		/* Move to first CG-RAM address */
	/* We have 8x8 bytes for our own chars */
	for (a=0; a<8; a++) {
		for (b=0; b<8; b++) {
			write_data(0);		/* At init we clear the CG-RAM */
		}
	}
	pos_cursor(dcol,dline);
}


void wr_cgram(char which, char *buf)
{
unsigned int a;

	if ((which < 0) || (which > 7))
		return;
	write_command(LCD_SETCGRAM + (8 * which));	/* Move to CG-RAM address (char# * 8) */
	/* We have 8x8 bytes for our own chars */
	for (a = 0; a < 8; a++) {
		write_data(buf[a]);		/* Write the 8 char-defining bytes */
	}
	USLEEP(200);
	pos_cursor(dcol,dline);			/* Bring us back to DD-RAM */
}


/*
 * Following functions are for tty like handling of the panel
 * i.e. set cursor position, clear display, home cursor, etc.
 */

/*
 * Sets cursor to certain position
 * col  = column [0-19]
 * line = line [0-3]
 */
void pos_cursor(unsigned char col, unsigned char line)
{
unsigned char offset=0;
unsigned char pos=0;

	/* some sanity checks */
	if (col > (LCD_COLS-1) ) {
		col=0;
		line++;
	}
	if (line > (LCD_LINES-1) ) line = 0;

	/* line address offsets (weird, aren't they?) */
	if (a_type==0) {
		switch (line) {
			case 0:
				offset=0;
				break;
			case 1:
				offset=64;
				break;
			case 2:
				offset=20;
				break;
			case 3:
				offset=84;
				break;
		}
	} else {
		switch (line) {
			case 0:
				offset=0;
				break;
			case 1:
				offset=0x20;
				break;
			case 2:
				offset=0x40;
				break;
			case 3:
				offset=0x60;
				break;
		}
	}

	/* Command for addressing is LCD_SETDDRAM so we add all this here */
	pos = LCD_SETDDRAM + offset + col;
	/* Update our private position counter */
	dline = line;
	dcol = col;
	/* now write it out */
	write_command(pos);
}


/*
 * All functions after here are for device driver specific purposes
 */

/*
 * Called when the device is open()ed
 */
#ifdef KERNEL20
static int lcd_open(struct inode *minode, struct file *file)
#endif
#ifdef KERNEL21
static int lcd_open(struct inode *minode, struct file *mfile)
#endif
{
	if (lcd_usage != 0)
		return -EBUSY;

	MOD_INC_USE_COUNT;
	lcd_usage = 1;
#ifdef DEBUG
	printk(KERN_DEBUG "LCD opened\n");
#endif

return(0);
}


/*
 * Called when the device is close()ed
 */
#ifdef KERNEL20
static void lcd_release(struct inode *minode, struct file *mfile)
#endif
#ifdef KERNEL21
int lcd_release(struct inode *minode, struct file *mfile)
#endif
{
	MOD_DEC_USE_COUNT;
	lcd_usage=0;
#ifdef DEBUG
	printk(KERN_DEBUG "LCD module closed\n");
#endif

#ifdef KERNEL21
return 0;
#endif
}


/*
 * Called if data is to be written out to the device
 */
#ifdef KERNEL20
int lcd_write_byte(struct inode *minode, struct file *mfile, const char *gdata, int length)
#endif
#ifdef KERNEL21
ssize_t lcd_write_byte(struct file *inode, const char *gdata, size_t length, loff_t *off_what)
#endif
{
int i,cnt;
static const char *data;
char c,cgenbuf[8];

	data=gdata;

	c=0;
	for (i=0; i<length; i++) {
#ifdef KERNEL20
		c=get_user((char *)(data+i));
#endif
#ifdef KERNEL21
		get_user(c,data+i);
#endif
#ifdef DEBUG
		printk(KERN_DEBUG "LCD: starting %d\n",i);
		printk(KERN_DEBUG "LCD: which is '%c'\n",c);
#endif
		if (c==10) { /* line feed */
			dline++;
			dcol=0;
			pos_cursor(dcol,dline);
		} else
		if (c==27) {
			if ((i+1)<length) {
				i+=1;
#ifdef KERNEL20
				c=get_user((char *)(data+i));
#endif
#ifdef KERNEL21
				get_user(c,(char *)(data+i));
#endif
				if (c=='c') {
					/* Clear display */
					write_command(LCD_CLEAR);
					dcol=0;
					dline=0;
				} else
				if (c=='h') {
					/* Cursor home */
					write_command(LCD_SETDDRAM);
					dcol=0;
					dline=0;
				} else
				/* BEGIN - ADDED by janosch@base2.de */
				if (c=='p') {
					/* Set Cursor Pos */
					if ((i+4)<length) {
						/* get position from input
						 * ESC-p-cc,l - where cc is two dec long col and l is line
						 */
						/* -get the first decimal of the column */
						i+=1;
#ifdef KERNEL20
						c=get_user((char *)(data+i));
#endif
#ifdef KERNEL21
						get_user(c,(char *)(data+i));
#endif
						if (c>=48 && c<=57) {
							dcol = 10 * (c-48);
						}
						/* -get the second decimal of the column */
						i+=1;
#ifdef KERNEL20
						c=get_user((char *)(data+i));
#endif
#ifdef KERNEL21
						get_user(c,(char *)(data+i));
#endif
						if (c>=48 && c<=57) {
							dcol = dcol + (c-48);
						}
						/* -skip the ',' */
						i+=1;
						/* -get the line number */
						i+=1;
#ifdef KERNEL20
						c=get_user((char *)(data+i));
#endif
#ifdef KERNEL21
						get_user(c,(char *)(data+i));
#endif
						if (c>=48 && c<=57) {
							dline = (c-48);
						}
						/* set the new cursor position on the display */
						write_command(LCD_SETDDRAM);
						pos_cursor(dcol,dline);
					}
				} else
				/* END - ADDED by janosch@base2.de */
				if (c=='K') {
					/* Cursor off */
					write_command(LCD_ONOFF | 0x04);
				} else
				if ((c=='k')||(c=='B')) {
					/* Cursor on and blink off */
					write_command(LCD_ONOFF | 0x04 | 0x02);
				} else
				if (c=='b') {
					/* Cursor and blink on */
					write_command(LCD_ONOFF | 0x04 | 0x02 | 0x01);
				} else
				if (c=='R') {
					/* Reset display */
					init_lcd();
					dcol=0;
					dline=0;
				} else
				if (c<8) {
#ifdef DEBUG
					printk(KERN_DEBUG "Def.char\n");
#endif
					if ((i+8)<length) {
						for (cnt=0; cnt<8; cnt++) {
#ifdef KERNEL20
							cgenbuf[cnt]=get_user((char *)(data+i+cnt+1));
#endif
#ifdef KERNEL21
							get_user(cgenbuf[cnt],(char *)(data+i+cnt+1));
#endif
						};
						wr_cgram(c,cgenbuf);
						i+=8;
					}
#ifdef DEBUG
					else
						printk(KERN_DEBUG "Def. too short\n");
#endif
				}
			}
		} else
		if (c==13) { /* carriage return */
			dcol=0;
			pos_cursor(dcol,dline);
		} else {
			write_data(c);
			dcol++;
			if (dcol>(LCD_COLS-1)) {
				dcol=0;
				dline++;
				pos_cursor(dcol,dline);
			}
		}
	}
return(length);
}


/*
 * Called if data is to be read
 */
#ifdef KERNEL20
int lcd_read_byte(struct inode *inode, struct file *mfile, char *data, int length)
#endif
#ifdef KERNEL21
ssize_t lcd_read_byte(struct file *inode, char *udata, size_t length, loff_t *loff_what)
#endif
{
#ifdef DEBUG
	printk(KERN_DEBUG "LCD: Trying to read readonly device\n");
#endif
return(EACCES);
}


/*
 * Called when module is unloaded
 */
void cleanup_module(void)
{
	release_region(LCD_ADDRESS,3);
	printk(KERN_INFO "LCD: I/O region freed\n");
	if (unregister_chrdev(LCD_MAJOR,"LCD panel"))
		printk(KERN_ERR "LCD: module cleanup failed\n");
	else
		printk(KERN_INFO "LCD: module cleanup succeeded\n");
}


struct file_operations lcd_fops=
{
#ifdef KERNEL24
	owner:		THIS_MODULE,
#endif
	read:		lcd_read_byte,
	write:		lcd_write_byte,
	open:		lcd_open,
	release:	lcd_release,
};

/*
 * First function called when module is inserted
 */
int init_module(void)
{
	printk(KERN_INFO "LCD device %s init:\n",LCD_MODULE_VERSION);

	/* Check IO-Address */
	if (io!=0)
		LCD_ADDRESS = io ; 

	/* Check for lines */
	if ( (lines==1) || (lines==2) || (lines==4) )
		LCD_LINES=lines;
 
	/* Check for cols */
	if ((cols==16) || (cols==20) || (cols==32) || (cols==40))
		LCD_COLS=cols;
 
	/* Check for t_short */
	if (t_short!=0)
		TIME_SHORT=t_short ; 

	/* Check for t_short */
	if (t_long!=0)
		TIME_LONG=t_long ; 

	if (register_chrdev(LCD_MAJOR,"LCD panel",&lcd_fops))
		printk(KERN_ERR "unable to get major 120 for LCD device\n");
	else {
		if (!check_region(LCD_ADDRESS,3)) {
			request_region(LCD_ADDRESS,3,"LCD panel");
			printk(KERN_INFO "  got 3 addresses from %#x\n",LCD_ADDRESS);
			printk(KERN_INFO "  Display is %d Lines with %d Cols\n",LCD_LINES,LCD_COLS);
			printk(KERN_INFO "  Timing is Short: %d, Long: %d\n",TIME_SHORT,TIME_LONG);
			lcd_usage = 0;
			init_lcd();
			init_cgram();
			dline = 0;
			dcol = 0;
		} else {
			release_region(LCD_ADDRESS,3);
			unregister_chrdev(LCD_MAJOR,"LCD panel");
			printk(KERN_ERR "  could not get 3 addresses from %#x\n",LCD_ADDRESS);
		}
	}

return 0;
}
